/***********************************************************************************************************************
 *
 * blueBill Resources - open source birding
 * Copyright (C) 2009-2011 by Tidalwave s.a.s. (http://www.tidalwave.it)
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * WWW: http://bluebill.tidalwave.it
 * SCM: https://java.net/hg/bluebill~resources-src
 *
 **********************************************************************************************************************/
package it.tidalwave.bluebill.factsheet.xenocanto;

import it.tidalwave.bluebill.taxonomy.Taxon;
import it.tidalwave.bluebill.taxonomy.TaxonomyManager;
import it.tidalwave.bluebill.taxonomy.elmo.ElmoTaxonomyVocabulary;
import it.tidalwave.bluebill.taxonomy.io.Importer;
import it.tidalwave.netbeans.util.Locator;
import it.tidalwave.observation.ObservationSet;
import it.tidalwave.openrdf.elmo.ElmoManagerThreadLocal;
import it.tidalwave.semantic.io.GraphMarshaller;
import it.tidalwave.semantic.io.GraphSerializer;
import it.tidalwave.semantic.io.impl.DefaultGraphSerializer;
import it.tidalwave.semantic.io.impl.GraphMarshallerImpl;
import it.tidalwave.util.Id;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.List;
import java.util.SortedSet;
import java.util.TreeSet;
import javax.annotation.Nonnull;
import org.openrdf.elmo.ElmoManager;
import org.openrdf.elmo.ElmoModule;
import org.openrdf.elmo.sesame.SesameManagerFactory;
import org.openrdf.model.Graph;
import org.openrdf.model.Statement;
import org.openrdf.model.impl.URIImpl;
import org.openrdf.repository.Repository;
import org.openrdf.repository.RepositoryException;
import org.openrdf.repository.RepositoryResult;
import org.openrdf.repository.sail.SailRepository;
import org.openrdf.sail.memory.MemoryStore;

/***********************************************************************************************************************
 *
 * @author  Fabrizio Giudici
 * @version $Id$
 *
 **********************************************************************************************************************/
public class TaxonUtilities
  {
    protected SesameManagerFactory smf;

    protected Repository repository;

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    public void initialize()
      throws RepositoryException
      {
        repository = new SailRepository(new MemoryStore());
        repository.initialize();
        final ElmoModule elmoModule = new ElmoModule();
        smf = new SesameManagerFactory(elmoModule, repository);
        final ElmoManager em = smf.createElmoManager();
        ElmoManagerThreadLocal.set(em);
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    public void close()
      {
        ElmoManagerThreadLocal.get().close();
        ElmoManagerThreadLocal.set(null);
        smf.close();
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    public void import_ (final @Nonnull File file)
      throws IOException
      {
        final TaxonomyManager taxonomyManager = Locator.find(TaxonomyManager.class);
        final Importer importer = taxonomyManager.as(Importer.class);
        importer.import_(repository, file);
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    public List<Taxon> getTaxa()
       throws RepositoryException
      {
        // FIXME: findTaxa() doesn't work - returned objects are not populated with attributes
//        for (final String displayName : Arrays.asList("AOU 7th edition", "EBNItalia 2003", "Roberts VII"))
//          {
//            it.tidalwave.bluebill.taxonomy.Taxonomy taxonomy = taxonomyManager.findTaxonomyByName(displayName, repository);
//
//            for (final it.tidalwave.bluebill.taxonomy.Taxon taxon : taxonomy.findTaxa().results())
//              {
//                  System.err.println(taxon.getScientificName());
//              }
//          }

        final RepositoryResult<Statement> statements =
                repository.getConnection().getStatements(null, new URIImpl(ElmoTaxonomyVocabulary.URI_SCIENTIFIC_NAME_ID), null, false);

        final SortedSet<String> ids = new TreeSet<String>();

        while (statements.hasNext())
          {
            final Statement statement = statements.next();
            ids.add(statement.getObject().stringValue());
          }

        final List<Taxon> taxa = new ArrayList<Taxon>();

        for (final String id : ids)
          {
            taxa.add(new PatchedTaxon(new Id(id)));
          }

        return taxa;
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    protected static void write (final @Nonnull ObservationSet observationSet,
                          final @Nonnull File targetFile,
                          final @Nonnull String mimeType)
      throws Exception
      {
        final GraphMarshaller marshaller = new GraphMarshallerImpl();
        final Graph graph = marshaller.marshal(observationSet);

        targetFile.getParentFile().mkdirs();
        final OutputStream os = new FileOutputStream(targetFile);
        final GraphSerializer serializer = new DefaultGraphSerializer();
        serializer.handleNamespace("rdf",  "http://www.w3.org/1999/02/22-rdf-syntax-ns#");
        serializer.handleNamespace("rdfs", "http://www.w3.org/2000/01/rdf-schema#");
        serializer.handleNamespace("owl",  "http://www.w3.org/2002/07/owl#");
        serializer.handleNamespace("skos", "http://www.w3.org/2004/02/skos/core#");
        serializer.handleNamespace("dc",   "http://purl.org/dc/elements/1.1/");
        serializer.handleNamespace("foaf", "http://xmlns.com/foaf/0.1/");
        serializer.handleNamespace("bb",   "http://www.tidalwave.it/rdf/observation/2010/07/01#");
        serializer.handleNamespace("xap",  "http://ns.adobe.com/xap/1.0/");
        serializer.handleNamespace("bbb",  "http://bluebill.tidalwave.it/rdf/observation/2011/06/26#");
        serializer.write(graph, os, mimeType);
        os.close();
      }
  }
