/***********************************************************************************************************************
 *
 * blueBill Resources - open source birding
 * Copyright (C) 2009-2011 by Tidalwave s.a.s. (http://www.tidalwave.it)
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * WWW: http://bluebill.tidalwave.it
 * SCM: https://java.net/hg/bluebill~resources-src
 *
 **********************************************************************************************************************/
package it.tidalwave.bluebill.factsheet.xenocanto;

import it.tidalwave.bluebill.taxonomy.Taxon;
import javax.annotation.Nonnull;
import java.util.List;
import java.util.Locale;
import java.io.File;
import it.tidalwave.observation.ObservationSet;
import it.tidalwave.semantic.io.MimeTypes;
import it.tidalwave.util.Id;
import java.net.InetSocketAddress;
import java.net.Proxy;
import lombok.extern.slf4j.Slf4j;

/***********************************************************************************************************************
 *
 * @author  Fabrizio Giudici
 * @version $Id$
 *
 **********************************************************************************************************************/
@Slf4j
public class XenoCantoGenerator
  {
    protected final File targetFolder;

    protected final TaxonUtilities taxonUtilities = new TaxonUtilities();

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    public static void main (final @Nonnull String ... args)
      throws Exception
      {
        final XenoCantoGenerator taxonomyConverter = new XenoCantoGenerator(args);
        taxonomyConverter.run();
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    private XenoCantoGenerator (final @Nonnull String ... args)
      {
        final String baseDir = args[0];
        targetFolder = new File(baseDir + "/target/");
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    public void run()
      throws Exception
      {
        // FIXME: redo using just TaxonUniqueIdManager, such as in CatalogueOfLifeImporter
        taxonUtilities.initialize();
        taxonUtilities.import_(new File(targetFolder, "AOU7th.rdf"));
        taxonUtilities.import_(new File(targetFolder, "BirdsOfIndia2009.rdf"));
        taxonUtilities.import_(new File(targetFolder, "EBNItalia2003.rdf"));
        taxonUtilities.import_(new File(targetFolder, "EBNItalia2011.rdf"));
        taxonUtilities.import_(new File(targetFolder, "RobertsVII.rdf"));

        createResources(taxonUtilities.getTaxa());

        taxonUtilities.close();
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    private void createResources (final @Nonnull List<Taxon> taxa)
      throws Exception
      {
        final XenoCantoObservationSetProvider xcObservationSetProvider = new XenoCantoObservationSetProvider();

        if ("true".equals(System.getProperty("it.tidalwave.bluebill.useLocalProxy")))
          {
            log.warn("USING A LOCAL PROXY");
            xcObservationSetProvider.setProxy(new Proxy(Proxy.Type.HTTP, new InetSocketAddress("127.0.0.1", 8080)));
          }

        final XenoCantoTaxaIdMapper xenoCantoTaxaIdMapper = new XenoCantoTaxaIdMapper();

        int count = 0;
        int processed = 0;

        for (final Taxon taxon : taxa)
          {
            count++;
            final Id taxonId = taxon.getScientificNameId();
            final String fileNameBase = taxonId.stringValue().replaceAll(":", "_");
            
            log.info("******************** {}/{}: {} - {} - {} - {}", new Object[] {
                     count,
                     taxa.size(),
                     "",
//                                 taxon.getDisplayName(Locale.ITALIAN),
                     taxon.getDisplayName(Locale.ENGLISH),
                     taxon.getScientificName(), fileNameBase });
            
            final File file = new File("target/test-artifacts/xenocanto-" + fileNameBase + ".n3");
        
            if (file.exists())
              {
                log.info(">>>> already processed");
                processed++;
                continue;
              }

            try
              {
                final String xenoCantoId = xenoCantoTaxaIdMapper.findXenoCantoId(taxon);

                if (xenoCantoId != null)
                  {
                    final ObservationSet observationSet = xcObservationSetProvider.createObservationSet(taxonId, taxon.getScientificName());
                    save(observationSet, fileNameBase);
                    processed++;
                  }
              }
            catch (Exception e)
              {
                log.warn("SHOULD NOT OCCUR: cannot translate {} {}", taxonId, taxon.getScientificName());
                log.warn("", e);
              }
          }

        log.info("Available resources for {} / {}", processed, taxa.size());
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    private void save (final @Nonnull ObservationSet observationSet, final @Nonnull String fileName)
      throws Exception
      {
        TaxonUtilities.write(observationSet, new File("target/test-artifacts/xenocanto-" + fileName + ".n3"),   MimeTypes.MIME_N3);
        TaxonUtilities.write(observationSet, new File("target/test-artifacts/xenocanto-" + fileName + ".xml"),  MimeTypes.MIME_RDF_XML);
      }
  }
