/***********************************************************************************************************************
 *
 * blueBill Resources - open source birding
 * Copyright (C) 2009-2011 by Tidalwave s.a.s. (http://www.tidalwave.it)
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * WWW: http://bluebill.tidalwave.it
 * SCM: https://java.net/hg/bluebill~resources-src
 *
 **********************************************************************************************************************/
package it.tidalwave.bluebill.factsheet.xenocanto;

import lombok.Getter;
import lombok.Setter;
import javax.annotation.Nonnull;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.text.DateFormat;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.io.InputStream;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import org.xml.sax.EntityResolver;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.jaxen.XPath;
import it.tidalwave.role.spi.DefaultDisplayable;
import it.tidalwave.role.spi.DefaultIdentifiable;
import it.tidalwave.util.Id;
import it.tidalwave.util.NotFoundException;
import it.tidalwave.semantic.Property;
import it.tidalwave.semantic.Type;
import it.tidalwave.observation.Cardinality;
import it.tidalwave.observation.Observable;
import it.tidalwave.observation.ObservationSet;
import it.tidalwave.observation.Observation.Builder;
import it.tidalwave.observation.Source;
import it.tidalwave.observation.simple.SimpleObservable;
import it.tidalwave.observation.simple.SimpleObservationSet;
import it.tidalwave.observation.simple.SimpleSource;
import it.tidalwave.observation.simple.rdf.ObservationVocabulary;
import it.tidalwave.xml.dom4j.Dom4jXmlParser;
import it.tidalwave.xml.jaxen.JaxenXPathProvider;
import it.tidalwave.xml.XmlParser;
import it.tidalwave.xml.XPathProvider;
import it.tidalwave.semantic.GenericEntity;
import it.tidalwave.mobile.media.Media;
import it.tidalwave.semantic.EntityWithProperties;
import it.tidalwave.util.Key;
import java.net.Proxy;
import lombok.Cleanup;
import lombok.extern.slf4j.Slf4j;
import static it.tidalwave.semantic.GenericEntity.*;
import static it.tidalwave.mobile.media.Media.*;
import static it.tidalwave.observation.Observer.Observer;
import static it.tidalwave.observation.Location.Location;

/***********************************************************************************************************************
 *
 * @author  Fabrizio Giudici
 * @version $Id$
 *
 **********************************************************************************************************************/
@Slf4j
public class XenoCantoObservationSetProvider
  {
    private static final String CC_BY_NC_ND_2_5 = "http://creativecommons.org/licenses/by-nc-nd/2.5/";
    
    private static final Id OBS_CALL_TYPE = new Id("http://bluebill.tidalwave.it/rdf/observation/2011/06/26#CallType");
    private static final Id XAP_RATING = new Id("http://ns.adobe.com/xap/1.0/Rating");

    private static final String XENO_CANTO_URI = "http://www.xeno-canto.org";
    private static final String XENO_CANTO_DISPLAY_NAME = "xeno-canto";
    public static final GenericEntity XENO_CANTO = namedEntity(new Id(XENO_CANTO_URI), XENO_CANTO_DISPLAY_NAME);

    // TODO: move elsewhere, only needed for tests
    /* package */ static final Id DUNLIN_ID = new Id("urn:lsid:catalogueoflife.org:taxon:eff82234-29c1-102b-9a4a-00304854f820:ac2010");
    /* package */ static final Id FLAMINGO_ID = new Id("urn:bluebill.tidalwave.it:taxon:5c6c5ef0-95cb-11df-a4c2-002332c672e6");
    /* package */ static final Id FRANKLINS_GULL_ID = new Id("urn:lsid:catalogueoflife.org:taxon:eff8a20e-29c1-102b-9a4a-00304854f820:ac2010");
    /* package */ static final Id PIED_BILLED_GREBE_ID = new Id("urn:lsid:catalogueoflife.org:taxon:efed6628-29c1-102b-9a4a-00304854f820:ac2010");
    /* package */ static final Id WHITE_BACKED_WOODPECKER_ID = new Id("urn:lsid:catalogueoflife.org:taxon:f3394ce8-29c1-102b-9a4a-00304854f820:ac2010");
    /* package */ static final Id PACIFIC_LOON_ID = new Id("urn:lsid:catalogueoflife.org:taxon:245b2cba-60a7-102d-be47-00304854f810:ac2010");

    private static final String XPATH_TR = "/html:html/html:body/html:table/html:tr/html:td/html:div/html:div/html:center/html:table/html:tr";
    private static final String XPATH_URL = "html:td[1]/html:a[2]/@href";
    private static final String XPATH_DURATION = "html:td[1]/text()[last()]";
    private static final String XPATH_AUTHOR = "html:td[2]";
    private static final String XPATH_DATE = "html:td[3]";
    private static final String XPATH_TIME = "html:td[4]";
    private static final String XPATH_COUNTRY = "html:td[5]";
    private static final String XPATH_PLACE = "html:td[6]/html:a/text()";
    private static final String XPATH_CALL_TYPE = "html:td[8]";
    private static final String XPATH_ID = "html:td[10]/html:a/text()";
////            td[9]/text() Notes

    private final Map<Id, String> xenoCantoUrlMapByTaxonId = new HashMap<Id, String>();

    @Nonnull
    private final XmlParser xmlParser;

    @Nonnull
    private final XPathProvider xPathProvider;

    @Getter @Setter
    private Proxy proxy = Proxy.NO_PROXY;

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    protected final EntityResolver entityResolver = new EntityResolver()
      {
        @Override @Nonnull
        public InputSource resolveEntity (final @Nonnull String publicId, final @Nonnull String systemId)
          throws SAXException, IOException
          {
            if ("http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd".equals(systemId))
              {
                return new InputSource(XmlParser.class.getResourceAsStream("/xhtml1-transitional.dtd"));
              }

            return null;
          }
      };

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    public XenoCantoObservationSetProvider()
      throws IOException
      {
        this(new Dom4jXmlParser(), new JaxenXPathProvider());
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    /* package */ XenoCantoObservationSetProvider (final @Nonnull XmlParser xmlParser,
                                                   final @Nonnull XPathProvider xPathProvider)
      throws IOException
      {
        this.xmlParser = xmlParser;
        this.xPathProvider = xPathProvider;
        this.xmlParser.setEntityResolver(entityResolver);

        loadXenoCantoIdMap();
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    private void loadXenoCantoIdMap()
      throws IOException
      {
        final InputStream is = getClass().getResourceAsStream("xeno-canto-table.txt");
        final BufferedReader br = new BufferedReader(new InputStreamReader(is, "UTF-8"));

        for (;;)
          {
            final String s = br.readLine();

            if (s == null)
              {
                break;
              }
            
            if (!s.startsWith("urn:"))
              {
                continue;  
              }

            final String[] split = s.split("=");
            xenoCantoUrlMapByTaxonId.put(new Id(split[0].trim()), "http://www.xeno-canto.org/browse.php?query=" + split[1].trim());
          }
      }

    /*******************************************************************************************************************
     *
     * FIXME: pass a single Taxon, not the id + displayName
     *
     ******************************************************************************************************************/
    @Nonnull
    public ObservationSet createObservationSet (final @Nonnull Id taxonId, final @Nonnull String displayName)
      throws IOException
      {
        log.info("createObservationSet({})", taxonId);

        try
          {
            final URL url = urlFor(taxonId);
            @Cleanup final InputStream is = url.openConnection(proxy).getInputStream();
            final ObservationSet result = createObservationSet(taxonId, displayName, is);
            is.close();
            return result;
          }
        catch (RuntimeException e)
          {
            throw e;
          }
        catch (IOException e)
          {
            throw e;
          }
        catch (Exception e)
          {
            throw new IOException(e.toString()); // Java 5 / Android compatibility
          }
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    /* package */ ObservationSet createObservationSet (final @Nonnull Id taxonId,
                                                       final @Nonnull String displayName,
                                                       final @Nonnull InputStream is)
      throws Exception
      {
        log.info("createObservationSet({}, {})", taxonId, is);

        final long xmlParsingTime = System.currentTimeMillis();
        final Object document = xmlParser.parse(new TidyFilterInputStream(new XenoCantoFilterInputStream(is)));
        log.info(">>>> XML parsed in {} msec", System.currentTimeMillis() - xmlParsingTime);

        final long xPathParsingTime = System.currentTimeMillis();
        final ObservationSet result = createObservationSet(taxonId, displayName, document, xmlParser);
        log.info(">>>> XPATH evaluated in {} msec", System.currentTimeMillis() - xPathParsingTime);

        return result;
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    private ObservationSet createObservationSet (final @Nonnull Id taxonId,
                                                 final @Nonnull String displayName,
                                                 final @Nonnull Object document,
                                                 final @Nonnull XmlParser parser)
      throws Exception
      {
        log.info("parseSounds({}, {})", document, parser);

        final XPath exprTr       = xPathProvider.createXPath(document, XPATH_TR);
        final XPath exprAuthor   = xPathProvider.createXPath(document, XPATH_AUTHOR);
        final XPath exprDate     = xPathProvider.createXPath(document, XPATH_DATE);
        final XPath exprTime     = xPathProvider.createXPath(document, XPATH_TIME);
        final XPath exprCountry  = xPathProvider.createXPath(document, XPATH_COUNTRY);
        final XPath exprPlace    = xPathProvider.createXPath(document, XPATH_PLACE);
        final XPath exprCallType = xPathProvider.createXPath(document, XPATH_CALL_TYPE);
        final XPath exprId       = xPathProvider.createXPath(document, XPATH_ID);
        final XPath exprUrl      = xPathProvider.createXPath(document, XPATH_URL);
        final XPath exprDuration = xPathProvider.createXPath(document, XPATH_DURATION);

        final ObservationSet observationSet = new SimpleObservationSet("foo:bar");

        final List<Object> trs = exprTr.selectNodes(document);

        final Source xenoCanto = new SimpleSource(new DefaultIdentifiable(new Id("http://www.xeno-canto.org")),
                                                  new DefaultDisplayable("xeno-canto", ""),
                                                  new Type(ObservationVocabulary.FOAF_ORGANIZATION),
                                                  new Property<String>(ObservationVocabulary.FOAF_HOME_PAGE, "http://www.xeno-canto.org"));
        final Observable species = new SimpleObservable(new DefaultIdentifiable(taxonId),
                                                        new DefaultDisplayable(displayName, ""));

        final Map<Id, Media> mediaMapById = new HashMap<Id, Media>();

        for (int i = 1; i < trs.size(); i++)
          {
            final Object tr = trs.get(i);

            final String authorName = exprAuthor.stringValueOf(tr).replaceAll("\n", " ").trim();
            final String dateAsString = (String)exprDate.stringValueOf(tr).replaceAll("\n", " ").trim();
            final String timeAsString = (String)exprTime.stringValueOf(tr).replaceAll("\n", " ").trim();
            final String country = (String)exprCountry.stringValueOf(tr).replaceAll("\n", " ").trim();
            final String place = (String)exprPlace.stringValueOf(tr).replaceAll("\n", " ").trim();
            final String location = String.format("%s (%s)", place, country);
            final String callType = (String)exprCallType.stringValueOf(tr).replaceAll("\n", " ").trim();
            final String id = (String)exprId.stringValueOf(tr).replaceAll("\n", " ").trim();
            final String url = (String)exprUrl.stringValueOf(tr).replaceAll("\n", " ").trim();
            final String durationAsString = (String)exprDuration.stringValueOf(tr).replaceAll("\n", " ").trim().replaceAll("\\(|\\)", "");
            final String durationSplit[] = durationAsString.split(":");

            final Id mediaId = new Id(url);
            Media sound = mediaMapById.get(mediaId);
            //
            // Make sure that a single instance of Media is used in case of multiple references to the same sound.
            // See BBMA-116.
            // While it would be a good idea to accumulate attributes, because sometimes repeated data is not
            // consistent (i.e. not all the repetitions have all the attributes), we can't as Media is immutable and
            // adding an attribute means to change the instance.
            //
            if (sound == null)
              {
                sound = new Media() //.with(RES_URL, new URI(URLEncoder.encode(url, "UTF-8")))
                                 .with(ID, mediaId)
                                 .with(MIME_TYPE, "audio/mp3")
                                 .with(RIGHTS, CC_BY_NC_ND_2_5);

                try
                  {
                    final int duration = Integer.parseInt(durationSplit[0]) * 60 + Integer.parseInt(durationSplit[1]);
                    sound = sound.with(DURATION, duration);
                  }
                catch (NumberFormatException e)
                  {
                    log.warn("Cannot parse duration: {}", e.toString());
                  }

                mediaMapById.put(mediaId, sound);
              }
            

            Builder observationBuilder = observationSet.createObservation()
                                            .of(species, Cardinality.UNDEFINED)
                                            .by(observationSet.findOrCreate(Observer, authorName))
                                            .at(observationSet.findOrCreate(Location, location))
                                            .from(xenoCanto)
                                            .with(new Id("http://www.xeno-canto.org/" + id.replaceAll("^XC", "")))
                                            .with(new DefaultDisplayable(callType, ""))
                                            .with(sound);

            observationBuilder = withRating(observationBuilder, callType);
            observationBuilder = withCallType(observationBuilder, callType);

            try
              {
                if (dateAsString != null)
                  {
                    observationBuilder = observationBuilder.at(parse(dateAsString));
                  }
              }
            catch (ParseException e)
              {
                log.warn("Cannot parse date: {}", e.toString());
              }

            observationBuilder.build();
          }

        return observationSet;
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull 
    private static Builder withRating (final @Nonnull Builder observationBuilder, final @Nonnull String callType) 
      {
        final String xcRating = callType.replaceAll("^.*\\((.)\\).*$", "$1").toUpperCase();

        if (xcRating.length() != 1)
          {
            log.warn("Invalid rating: {} - from: {}", xcRating, callType);
          }
        else
          {
            final int rating = 5 - (xcRating.charAt(0) - 'A');  

            if ((rating < 0) || (rating > 5))
              {
                log.warn("Invalid rating: {} - from: {}", rating, callType);
              }
            else if (rating > 0)
              {
                log.info(">>>> rating: {} - {}", xcRating, rating);
                return observationBuilder.with(new Property<Integer>(XAP_RATING, rating));
              }
          }
        
        return observationBuilder;
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull 
    private static Builder withCallType (@Nonnull Builder observationBuilder, final @Nonnull String callType) 
      {
        final String c = callType.replaceAll("\\(.\\)", "").trim().toLowerCase().replaceAll("\\.$", "").replace("(no score)", "");
        
        for (final String cc : c.replace(" and ", "&").replace(";", "&").split("&"))
          {
            String ccc = cc.trim();
            // TODO: if contains ? add uncertainty
            ccc = ccc.replace("?", "").trim();
            log.info("call type: {}", ccc);            
            observationBuilder = observationBuilder.with(new Property<String>(OBS_CALL_TYPE, ccc));
          }
        
        return observationBuilder;
      }
    
    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    /* package */ URL urlFor (final @Nonnull Id id)
      throws NotFoundException
      {
        try
          {
            return new URL(NotFoundException.throwWhenNull(xenoCantoUrlMapByTaxonId.get(id), "No URL for %s", id));
          }
        catch (MalformedURLException e)
          {
            throw new RuntimeException(e);
          }
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    protected Date parse (@Nonnull String string)
      throws ParseException
      {
        string = string.replace("x", "?").replace("xx", "?").replace("??", "?");
        final DateFormat[] df = new DateFormat[]
          {
            new SimpleDateFormat("dd-MM-yy"),
            new SimpleDateFormat("dd/MM/yy"),
            new SimpleDateFormat("dd-MM-yyyy"),
            new SimpleDateFormat("dd -MM-yyyy"),
            new SimpleDateFormat("dd -MM - yyyy"),
            new SimpleDateFormat("dd - MM - yyyy"),
            new SimpleDateFormat("dd_MM_yyyy"),
            new SimpleDateFormat("dd MM yyyy"),
            new SimpleDateFormat("d MM yyyy"),
            new SimpleDateFormat("dd MM yy"),
            new SimpleDateFormat("dd MMM yy"),
            new SimpleDateFormat("dd MMM yyyy"),
            new SimpleDateFormat("dd MMMM yy"),
            new SimpleDateFormat("d MMMM yy"),
            new SimpleDateFormat("dd MMMMMMMMMMMMM yyyy"),
            new SimpleDateFormat("dd.MM.yyyy"),
            new SimpleDateFormat("ddMMyyyy"),
            new SimpleDateFormat("ddMMyy"),
            new SimpleDateFormat("MMM-dd-yy"),
            new SimpleDateFormat("MMM-d-yy"),
            new SimpleDateFormat("MMM dd, yyyy"),
            new SimpleDateFormat("MMMMMMMMMMMMM dd, yyyy"),
            new SimpleDateFormat("MMMMMMMMMMMMM d, yyyy")
          };

        for (final DateFormat d : df)
          {
            try
              {
                return d.parse(string);
              }
            catch (ParseException e)
              {
                // try next
              }
          }

        throw new ParseException(string, 0);
      }
  }
