/***********************************************************************************************************************
 *
 * blueBill Resources - open source birding
 * Copyright (C) 2009-2011 by Tidalwave s.a.s. (http://www.tidalwave.it)
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * WWW: http://bluebill.tidalwave.it
 * SCM: https://java.net/hg/bluebill~resources-src
 *
 **********************************************************************************************************************/
package it.tidalwave.bluebill.factsheet.xenocanto;

import it.tidalwave.bluebill.taxonomy.Taxon;
import javax.annotation.Nonnull;
import java.util.HashMap;
import java.util.Map;
import java.util.zip.GZIPInputStream;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import it.tidalwave.util.Id;
import it.tidalwave.util.NotFoundException;
import java.io.InputStream;
import lombok.extern.slf4j.Slf4j;

/***********************************************************************************************************************
 *
 * @author  Fabrizio Giudici
 * @version $Id$
 *
 **********************************************************************************************************************/
@Slf4j
public class XenoCantoTaxaIdMapper   
  {
    private final Map<String, String> xenoCantoIdMapByScientificName;
    private final Map<String, String> lsidScientificNameMapById;

    private final Map<String, String> lsidToXenoCantoScientificNameMap;

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    public XenoCantoTaxaIdMapper()
      throws IOException, Exception
      {
        xenoCantoIdMapByScientificName = loadXenoCantoIdMap();
        lsidScientificNameMapById = loadLSIDMap();
        lsidToXenoCantoScientificNameMap = loadLsidToXenoCantoScientificNameMap();
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    private static Map<String, String> loadLsidToXenoCantoScientificNameMap()
      throws IOException
      {
        final InputStream is = XenoCantoTaxaIdMapper.class.getResourceAsStream("xenocanto.config");
        final BufferedReader br = new BufferedReader(new InputStreamReader(is, "UTF-8"));
        final Map<String, String> result = new HashMap<String, String>();

        for (;;)
          {
            String s = br.readLine();

            if (s == null)
              {
                break;
              }

            s = s.replace("#.*", "").trim();

            if (s.equals(""))
              {
                continue;
              }

            s = s.replaceAll("->", "#");
            final String[] split = s.split("#");
//            logger.fine("%s -> %s\n", split[0], split[2]);
            result.put(split[1].trim(), split[0].trim());
          }

        return result;
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    private static Map<String, String> loadXenoCantoIdMap()
      throws IOException
      {
        final File file = new File("src/test/resources/all_species.txt");
        final BufferedReader br = new BufferedReader(new InputStreamReader(new FileInputStream(file), "UTF-8"));
        final Map<String, String> result = new HashMap<String, String>();

        for (;;)
          {
            final String s = br.readLine();

            if (s == null)
              {
                break;
              }

            final String[] split = s.split("\\$");
//            logger.fine("%s -> %s\n", split[0], split[2]);
            result.put(split[0], split[2]);
          }

        return result;
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    private static Map<String, String> loadLSIDMap()
      throws IOException
      {
        final File file = new File("target/TaxonIdMap.txt.gz");
        final BufferedReader br = new BufferedReader(new InputStreamReader(new GZIPInputStream(new FileInputStream(file)), "UTF-8"));
        final Map<String, String> result = new HashMap<String, String>();

        for (;;)
          {
            final String s = br.readLine();

            if (s == null)
              {
                break;
              }

            final int i = s.indexOf(':');
            final String path = s.substring(0, i).trim().replaceAll("^.*dae/", "").replaceAll("^.*nae/", "").replace('/', ' ');
            final String id = s.substring(i + 1).trim();
            log.info("LSID {} -> {}", id, path);
            result.put(id, path);
          }

        return result;
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    public String findXenoCantoId (final @Nonnull Taxon taxon)
      throws NotFoundException
      {
        final Id id = taxon.getScientificNameId();
        final String lsidScientificName = lsidScientificNameMapById.get(id.stringValue());
        String xenoCantoScientificName = lsidToXenoCantoScientificNameMap.get(lsidScientificName);
        xenoCantoScientificName = (xenoCantoScientificName != null) ? xenoCantoScientificName : lsidScientificName;
        final String xenoCantoId = xenoCantoIdMapByScientificName.get(xenoCantoScientificName);
        log.info("TRANSLATE: id: {} -> LSID scientific name: {} -> xenoCantoScientificName: {} xenoCantoId: {}", 
                 new Object[] { id, lsidScientificName, xenoCantoScientificName, xenoCantoId });

        return xenoCantoId;
      }
  }
